@@ -13,6 +13,7 @@ pub struct Consensus<T: Proposition> {
     pub elders: PublicKeySet,
     pub n_elders: usize,
     pub secret_key: (NodeId, SecretKeyShare),
+    pub processed_votes_cache: BTreeSet<SignatureShare>,
     pub votes: BTreeMap<NodeId, SignedVote<T>>,
     pub faults: BTreeMap<NodeId, Fault<T>>,
     pub decision: Option<Decision<T>>,
@@ -25,6 +26,12 @@ pub struct Decision<T: Proposition> {
     pub faults: BTreeSet<Fault<T>>,
 }
 
+impl<T: Proposition> Decision<T> {
+    pub fn faulty_ids(&self) -> BTreeSet<NodeId> {
+        BTreeSet::from_iter(self.faults.iter().map(Fault::voter_at_fault))
+    }
+}
+
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub enum VoteResponse<T: Proposition> {
     WaitingForMoreVotes,
@@ -41,6 +48,7 @@ impl<T: Proposition> Consensus<T> {
             elders,
             n_elders,
             secret_key,
+            processed_votes_cache: Default::default(),
             votes: Default::default(),
             faults: Default::default(),
             decision: None,
@@ -59,6 +67,10 @@ impl<T: Proposition> Consensus<T> {
         BTreeSet::from_iter(self.faults.values().cloned())
     }
 
+    pub fn faulty_ids(&self) -> BTreeSet<NodeId> {
+        BTreeSet::from_iter(self.faults.keys().copied())
+    }
+
     pub fn build_super_majority_vote(
         &self,
         votes: BTreeSet<SignedVote<T>>,
@@ -66,15 +78,20 @@ impl<T: Proposition> Consensus<T> {
         gen: Generation,
     ) -> Result<SignedVote<T>> {
         let faulty = BTreeSet::from_iter(faults.iter().map(Fault::voter_at_fault));
-        let proposals: BTreeMap<T, (NodeId, SignatureShare)> =
-            crate::vote::proposals(&votes, &faulty)
-                .into_iter()
-                .map(|p| {
-                    let sig = self.sign(&p)?;
-                    Ok((p, (self.id(), sig)))
-                })
-                .collect::<Result<_>>()?;
+
+        let proposals = VoteCount::count(&votes, &faulty)
+            .candidate_with_most_votes()
+            .map(|(candidate, _)| candidate.proposals.clone())
+            .unwrap_or_default()
+            .into_iter()
+            .map(|proposal| {
+                let sig = self.sign(&proposal)?;
+                Ok((proposal, (self.id(), sig)))
+            })
+            .collect::<Result<_>>()?;
+
         let ballot = Ballot::SuperMajority { votes, proposals }.simplify();
+
         let vote = Vote {
             gen,
             ballot,
@@ -83,64 +100,47 @@ impl<T: Proposition> Consensus<T> {
         self.sign_vote(vote)
     }
 
-    pub fn is_new_vote_from_voter(&self, signed_vote: &SignedVote<T>) -> bool {
-        if let Some(previous_vote_from_voter) = self.votes.get(&signed_vote.voter) {
-            signed_vote.strict_supersedes(previous_vote_from_voter)
-        } else {
-            // if we have no votes from this voter, then it is new
-            true
-        }
-    }
-
     // handover: gen = gen
     // membership: gen = pending_gen
     /// Handles a signed vote
     /// Returns the vote we cast and the reached consensus vote in case consensus was reached
     pub fn handle_signed_vote(&mut self, signed_vote: SignedVote<T>) -> Result<VoteResponse<T>> {
-        info!("[MBR-{}] handling vote {:?}", self.id(), signed_vote);
+        info!("[{}] handling vote {:?}", self.id(), signed_vote);
+
+        if self.decision.is_some() {
+            info!("[{}] we've decided already, dropping vote", self.id());
+            return Ok(VoteResponse::WaitingForMoreVotes);
+        }
+
+        if self.have_we_processed_vote(&signed_vote) {
+            info!("[{}] dropping already processed vote", self.id());
+            return Ok(VoteResponse::WaitingForMoreVotes);
+        }
+
+        self.validate_signed_vote(&signed_vote)?;
+
         if let Err(faults) = self.detect_byzantine_voters(&signed_vote) {
-            info!("[MBR-{}] Found faults {:?}", self.id(), faults);
+            info!("[{}] Found faults {:?}", self.id(), faults);
             self.faults.extend(faults);
         }
+
         if self.faults.contains_key(&signed_vote.voter) {
-            info!("[MBR-{}] dropping vote from faulty voter", self.id());
-            return Ok(VoteResponse::WaitingForMoreVotes);
-        }
-        if !self.is_new_vote_from_voter(&signed_vote) {
-            info!("[MBR-{}] dropping old vote {:?}", self.id(), signed_vote);
+            info!("[{}] dropping vote from faulty voter", self.id());
             return Ok(VoteResponse::WaitingForMoreVotes);
         }
 
-        let their_decision = self.get_decision(&signed_vote.vote_count())?;
+        self.process_signed_vote(signed_vote)
+    }
 
-        if let Some(decision) = self.decision.clone() {
-            let resp = if their_decision.is_none() {
-                info!(
-                    "[MBR-{}] We've already terminated, responding with decision",
-                    self.id()
-                );
-                let vote = self.build_super_majority_vote(
-                    decision.votes,
-                    decision.faults,
-                    signed_vote.vote.gen,
-                )?;
-                VoteResponse::Broadcast(vote)
-            } else {
-                info!("[MBR-{}] We've both terminated, no-op", self.id());
-                VoteResponse::WaitingForMoreVotes
-            };
+    fn process_signed_vote(&mut self, signed_vote: SignedVote<T>) -> Result<VoteResponse<T>> {
+        self.log_processed_signed_vote(&signed_vote);
 
-            return Ok(resp);
-        }
-
-        self.log_signed_vote(&signed_vote);
-
-        if let Some(proposals) = their_decision {
+        if let Some(proposals) = self.get_decision(&signed_vote.vote_count())? {
             // This case is here to handle situations where this node has recieved
             // a faulty vote previously that is preventing it from accepting a network
             // decision using the sm_over_sm logic below.
             info!(
-                "[MBR-{}] They terminated but we haven't yet, accepting decision",
+                "[{}] They terminated but we haven't yet, accepting decision",
                 self.id()
             );
             let votes = crate::vote::simplify_votes(&self.votes.values().cloned().collect());
@@ -153,14 +153,11 @@ impl<T: Proposition> Consensus<T> {
             return Ok(VoteResponse::WaitingForMoreVotes);
         }
 
-        let vote_count = VoteCount::count(
-            self.votes.values(),
-            &BTreeSet::from_iter(self.faults.keys().copied()),
-        );
+        let vote_count = VoteCount::count(self.votes.values(), &self.faulty_ids());
 
         if let Some(proposals) = self.get_decision(&vote_count)? {
             info!(
-                "[MBR-{}] Detected super majority over super majorities: {proposals:?}",
+                "[{}] Detected super majority over super majorities: {proposals:?}",
                 self.id()
             );
             let votes = crate::vote::simplify_votes(&self.votes.values().cloned().collect());
@@ -169,12 +166,17 @@ impl<T: Proposition> Consensus<T> {
                 proposals,
                 faults: self.faults(),
             };
+            let vote = self.build_super_majority_vote(
+                decision.votes.clone(),
+                decision.faults.clone(),
+                signed_vote.vote.gen,
+            )?;
             self.decision = Some(decision);
-            return Ok(VoteResponse::WaitingForMoreVotes);
+            return Ok(VoteResponse::Broadcast(vote));
         }
 
         if self.is_split_vote(&vote_count) {
-            info!("[MBR-{}] Detected split vote", self.id());
+            info!("[{}] Detected split vote", self.id());
             let merge_vote = Vote {
                 gen: signed_vote.vote.gen,
                 ballot: Ballot::Merge(self.votes.values().cloned().collect()).simplify(),
@@ -183,10 +185,10 @@ impl<T: Proposition> Consensus<T> {
             let signed_merge_vote = self.sign_vote(merge_vote)?;
 
             let resp = if vote_count != signed_merge_vote.vote_count() {
-                info!("[MBR-{}] broadcasting merge.", self.id());
-                VoteResponse::Broadcast(self.cast_vote(&signed_merge_vote)?)
+                info!("[{}] broadcasting merge.", self.id());
+                VoteResponse::Broadcast(self.cast_vote(signed_merge_vote)?)
             } else {
-                info!("[MBR-{}] merge does not change counts, waiting.", self.id());
+                info!("[{}] merge does not change counts, waiting.", self.id());
                 VoteResponse::WaitingForMoreVotes
             };
 
@@ -194,25 +196,25 @@ impl<T: Proposition> Consensus<T> {
         }
 
         if self.is_super_majority(&vote_count) {
-            info!("[MBR-{}] Detected super majority", self.id());
+            info!("[{}] Detected super majority", self.id());
 
             if let Some(our_vote) = self.votes.get(&self.id()) {
                 // We voted during this generation.
 
                 if our_vote.vote.is_super_majority_ballot() {
-                    info!("[MBR-{}] We've already sent a super majority, waiting till we either have a split vote or SM / SM", self.id());
+                    info!("[{}] We've already sent a super majority, waiting till we either have a split vote or SM / SM", self.id());
                     return Ok(VoteResponse::WaitingForMoreVotes);
                 }
             }
 
-            info!("[MBR-{}] broadcasting super majority", self.id());
+            info!("[{}] broadcasting super majority", self.id());
             let signed_vote = self.build_super_majority_vote(
                 self.votes.values().cloned().collect(),
                 BTreeSet::from_iter(self.faults.values().cloned()),
                 signed_vote.vote.gen,
             )?;
 
-            return Ok(VoteResponse::Broadcast(self.cast_vote(&signed_vote)?));
+            return Ok(VoteResponse::Broadcast(self.cast_vote(signed_vote)?));
         }
 
         // We have determined that we don't yet have enough votes to take action.
@@ -224,14 +226,14 @@ impl<T: Proposition> Consensus<T> {
                 faults: self.faults(),
             })?;
             info!(
-                "[MBR-{}] adopting ballot {:?}",
+                "[{}] adopting ballot {:?}",
                 self.id(),
                 signed_vote.vote.ballot
             );
 
-            Ok(VoteResponse::Broadcast(self.cast_vote(&signed_vote)?))
+            Ok(VoteResponse::Broadcast(self.cast_vote(signed_vote)?))
         } else {
-            info!("[MBR-{}] waiting for more votes", self.id());
+            info!("[{}] waiting for more votes", self.id());
             Ok(VoteResponse::WaitingForMoreVotes)
         }
     }
@@ -244,18 +246,25 @@ impl<T: Proposition> Consensus<T> {
         })
     }
 
-    pub fn cast_vote(&mut self, signed_vote: &SignedVote<T>) -> Result<SignedVote<T>> {
+    pub fn cast_vote(&mut self, signed_vote: SignedVote<T>) -> Result<SignedVote<T>> {
+        info!("[{}] casting vote {:?}", self.id(), signed_vote);
         match self.handle_signed_vote(signed_vote.clone())? {
-            VoteResponse::WaitingForMoreVotes => Ok(signed_vote.clone()),
+            VoteResponse::WaitingForMoreVotes => Ok(signed_vote),
             VoteResponse::Broadcast(vote) => Ok(vote),
         }
     }
 
-    pub fn log_signed_vote(&mut self, signed_vote: &SignedVote<T>) {
+    fn have_we_processed_vote(&self, signed_vote: &SignedVote<T>) -> bool {
+        self.processed_votes_cache.contains(&signed_vote.sig)
+    }
+
+    fn log_processed_signed_vote(&mut self, signed_vote: &SignedVote<T>) {
         for vote in signed_vote.unpack_votes() {
-            let existing_vote = self.votes.entry(vote.voter).or_insert_with(|| vote.clone());
-            if vote.supersedes(existing_vote) {
-                *existing_vote = vote.clone()
+            if self.processed_votes_cache.insert(vote.sig.clone()) {
+                let existing_vote = self.votes.entry(vote.voter).or_insert_with(|| vote.clone());
+                if vote.supersedes(existing_vote) {
+                    *existing_vote = vote.clone()
+                }
             }
         }
     }
@@ -283,22 +292,12 @@ impl<T: Proposition> Consensus<T> {
     }
 
     fn get_decision(&self, vote_count: &VoteCount<T>) -> Result<Option<BTreeMap<T, Signature>>> {
-        if let Some((_candidate, shares_by_voter)) = vote_count.super_majority_with_most_votes() {
-            if shares_by_voter.len() > self.elders.threshold() {
-                let mut proposal_sigs: BTreeMap<T, BTreeSet<(u64, SignatureShare)>> =
-                    Default::default();
-
-                for (id, shares) in shares_by_voter {
-                    shares.iter().for_each(|(prop, sig)| {
-                        proposal_sigs
-                            .entry(prop.clone())
-                            .or_default()
-                            .insert((*id as u64, sig.clone()));
-                    });
-                }
-                let proposals = proposal_sigs
-                    .into_iter()
-                    .map(|(prop, sigs)| Ok((prop, self.elders.combine_signatures(sigs)?)))
+        if let Some((_candidate, sm_count)) = vote_count.super_majority_with_most_votes() {
+            if sm_count.count > self.elders.threshold() {
+                let proposals = sm_count
+                    .proposals
+                    .iter()
+                    .map(|(prop, sigs)| Ok((prop.clone(), self.elders.combine_signatures(sigs)?)))
                     .collect::<Result<_>>()?;
                 return Ok(Some(proposals));
             }
@@ -312,28 +311,32 @@ impl<T: Proposition> Consensus<T> {
         signed_vote: &SignedVote<T>,
     ) -> std::result::Result<(), BTreeMap<NodeId, Fault<T>>> {
         let mut faults = BTreeMap::new();
-        if let Some(existing_vote) = self.votes.get(&signed_vote.voter) {
-            let fault = Fault::ChangedVote {
-                a: existing_vote.clone(),
-                b: signed_vote.clone(),
-            };
+        for vote in signed_vote.unpack_votes() {
+            if self.have_we_processed_vote(vote) {
+                continue;
+            }
 
-            if let Ok(()) = fault.validate(&self.elders) {
-                faults.insert(signed_vote.voter, fault);
+            if let Some(existing_vote) = self.votes.get(&vote.voter) {
+                let fault = Fault::ChangedVote {
+                    a: existing_vote.clone(),
+                    b: vote.clone(),
+                };
+
+                if let Ok(()) = fault.validate(&self.elders) {
+                    faults.insert(vote.voter, fault);
+                }
+            }
+
+            {
+                let fault = Fault::InvalidFault {
+                    signed_vote: vote.clone(),
+                };
+                if let Ok(()) = fault.validate(&self.elders) {
+                    faults.insert(vote.voter, fault);
+                }
             }
         }
 
-        {
-            let fault = Fault::InvalidFault {
-                signed_vote: signed_vote.clone(),
-            };
-            if let Ok(()) = fault.validate(&self.elders) {
-                faults.insert(signed_vote.voter, fault);
-            }
-        }
-
-        // TODO: recursively check for faulty votes
-
         if faults.is_empty() {
             Ok(())
         } else {
@@ -343,7 +346,7 @@ impl<T: Proposition> Consensus<T> {
 
     /// Validates a vote recursively all the way down to the proposition (T)
     /// Assumes those propositions are correct, they MUST be checked beforehand by the caller
-    pub fn validate_signed_vote(&self, signed_vote: &SignedVote<T>) -> Result<()> {
+    fn validate_signed_vote(&self, signed_vote: &SignedVote<T>) -> Result<()> {
         signed_vote.validate_signature(&self.elders)?;
         self.validate_vote(&signed_vote.vote)?;
         Ok(())
@@ -352,24 +355,19 @@ impl<T: Proposition> Consensus<T> {
     fn validate_vote(&self, vote: &Vote<T>) -> Result<()> {
         match &vote.ballot {
             Ballot::Propose(_) => Ok(()),
-            Ballot::Merge(votes) => {
-                // TODO: generation checking needs to move to sn_membership
-                for child_vote in votes.iter() {
-                    if child_vote.vote.gen != vote.gen {
-                        return Err(Error::MergedVotesMustBeFromSameGen {
-                            child_gen: child_vote.vote.gen,
-                            merge_gen: vote.gen,
-                        });
-                    }
-                    self.validate_signed_vote(child_vote)?;
-                }
-                Ok(())
-            }
+            Ballot::Merge(votes) => self.validate_child_votes(vote.gen, votes),
             Ballot::SuperMajority { votes, proposals } => {
-                if !self.is_super_majority(&VoteCount::count(votes, &vote.known_faulty())) {
+                let vote_count = VoteCount::count(votes, &vote.faulty_ids());
+
+                let candidate_proposals = vote_count
+                    .candidate_with_most_votes()
+                    .map(|(c, _)| c.proposals.clone())
+                    .unwrap_or_default();
+
+                if !self.is_super_majority(&vote_count) {
                     // TODO: this should be moved to fault detection
                     Err(Error::SuperMajorityBallotIsNotSuperMajority)
-                } else if vote.proposals() != BTreeSet::from_iter(proposals.keys().cloned()) {
+                } else if !candidate_proposals.iter().eq(proposals.keys()) {
                     // TODO: this should be moved to fault detection
                     Err(Error::SuperMajorityProposalsDoesNotMatchVoteProposals)
                 } else if proposals
@@ -381,21 +379,27 @@ impl<T: Proposition> Consensus<T> {
                 {
                     Err(Error::InvalidElderSignature)
                 } else {
-                    for child_vote in votes.iter() {
-                        // TODO: generation checking needs to move to sn_membership
-                        if child_vote.vote.gen != vote.gen {
-                            return Err(Error::MergedVotesMustBeFromSameGen {
-                                child_gen: child_vote.vote.gen,
-                                merge_gen: vote.gen,
-                            });
-                        }
-                        self.validate_signed_vote(child_vote)?;
-                    }
-                    Ok(())
+                    self.validate_child_votes(vote.gen, votes)
                 }
             }
         }
     }
+
+    fn validate_child_votes(&self, gen: Generation, votes: &BTreeSet<SignedVote<T>>) -> Result<()> {
+        for child_vote in votes {
+            // TODO: generation checking needs to move to sn_membership
+            if child_vote.vote.gen != gen {
+                return Err(Error::MergedVotesMustBeFromSameGen {
+                    child_gen: child_vote.vote.gen,
+                    merge_gen: gen,
+                });
+            }
+            if !self.have_we_processed_vote(child_vote) {
+                self.validate_signed_vote(child_vote)?
+            };
+        }
+        Ok(())
+    }
 }
 
 #[cfg(test)]
@@ -405,58 +409,58 @@ mod tests {
     use rand::{prelude::StdRng, SeedableRng};
 
     #[test]
-    fn test_is_new_vote_from_voter() {
+    fn test_have_we_seen_this_vote_before() {
         let mut rng = StdRng::from_seed([0u8; 32]);
         let elders_sk = SecretKeySet::random(10, &mut rng);
-        let mut consensus = Consensus::from(
-            (1u8, elders_sk.secret_key_share(1usize)),
-            elders_sk.public_keys(),
-            10,
-        );
-
-        consensus.votes = BTreeMap::from_iter((0..10u8).map(|i| {
-            (
-                i,
-                consensus
-                    .sign_vote(Vote {
-                        gen: 0,
-                        ballot: Ballot::Propose(i),
-                        faults: Default::default(),
-                    })
-                    .unwrap(),
+        let mut states = Vec::from_iter((1..=10).into_iter().map(|id| {
+            Consensus::from(
+                (id, elders_sk.secret_key_share(id as usize)),
+                elders_sk.public_keys(),
+                10,
             )
         }));
 
+        for i in 0..10u8 {
+            let vote = states[i as usize]
+                .sign_vote(Vote {
+                    gen: 0,
+                    ballot: Ballot::Propose(i),
+                    faults: Default::default(),
+                })
+                .unwrap();
+            states[0].log_processed_signed_vote(&vote);
+        }
+
         // try existing vote
-        let new_vote = consensus
+        let new_vote = states[2]
             .sign_vote(Vote {
                 gen: 0,
                 ballot: Ballot::Propose(2u8),
                 faults: Default::default(),
             })
             .unwrap();
-        assert!(!consensus.is_new_vote_from_voter(&new_vote));
+        assert!(states[0].have_we_processed_vote(&new_vote));
 
         // try merge vote superseding existing vote
-        let new_vote = consensus
+        let new_vote = states[0]
             .sign_vote(Vote {
                 gen: 0,
                 ballot: Ballot::Merge(BTreeSet::from_iter(
-                    consensus.votes.iter().map(|(_, v)| v.clone()),
+                    states[0].votes.iter().map(|(_, v)| v.clone()),
                 )),
                 faults: Default::default(),
             })
             .unwrap();
-        assert!(consensus.is_new_vote_from_voter(&new_vote));
+        assert!(!states[0].have_we_processed_vote(&new_vote));
 
         // try bad vote not superseding existing
-        let new_vote = consensus
+        let new_vote = states[0]
             .sign_vote(Vote {
                 gen: 0,
                 ballot: Ballot::Propose(44u8),
                 faults: Default::default(),
             })
             .unwrap();
-        assert!(!consensus.is_new_vote_from_voter(&new_vote));
+        assert!(!states[0].have_we_processed_vote(&new_vote));
     }
 }
