@@ -73,12 +73,34 @@ pub struct Candidate<T> {
     pub faulty: BTreeSet<NodeId>,
 }
 
-type SignatureSharesByVoter<T> = BTreeMap<NodeId, BTreeMap<T, SignatureShare>>;
+impl<T> Default for Candidate<T> {
+    fn default() -> Self {
+        Self {
+            proposals: Default::default(),
+            faulty: Default::default(),
+        }
+    }
+}
+
+#[derive(Debug, Clone, PartialEq, Eq)]
+pub struct SuperMajorityCount<T> {
+    pub count: usize,
+    pub proposals: BTreeMap<T, BTreeMap<u64, SignatureShare>>,
+}
+
+impl<T> Default for SuperMajorityCount<T> {
+    fn default() -> Self {
+        Self {
+            count: 0,
+            proposals: Default::default(),
+        }
+    }
+}
 
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub struct VoteCount<T> {
     pub candidates: BTreeMap<Candidate<T>, usize>,
-    pub super_majorities: BTreeMap<Candidate<T>, SignatureSharesByVoter<T>>,
+    pub super_majorities: BTreeMap<Candidate<T>, SuperMajorityCount<T>>,
     pub voters: BTreeSet<NodeId>,
 }
 
@@ -98,41 +120,50 @@ impl<T: Proposition> VoteCount<T> {
         faulty: &BTreeSet<NodeId>,
     ) -> Self {
         let mut count: VoteCount<T> = VoteCount::default();
-        let mut counted: BTreeSet<SignatureShare> = Default::default();
+
+        let mut votes_by_honest_voter: BTreeMap<NodeId, SignedVote<T>> = Default::default();
 
         for vote in votes.into_iter() {
             for unpacked_vote in vote.borrow().unpack_votes() {
-                // We always include voters in the voter set (even if they are faulty)
-                // so that we have an accurate reading of who has contributed votes.
-                // This is done to to aid in split-vote detection
-                count.voters.insert(unpacked_vote.voter);
-
-                if counted.contains(&unpacked_vote.sig) || faulty.contains(&unpacked_vote.voter) {
+                if faulty.contains(&unpacked_vote.voter) {
                     continue;
                 }
-
-                counted.insert(unpacked_vote.sig.clone());
-
-                let candidate = unpacked_vote.vote.candidate();
-
-                match &unpacked_vote.vote.ballot {
-                    Ballot::SuperMajority { proposals, .. } => {
-                        let shares = count.super_majorities.entry(candidate).or_default();
-                        for (t, (id, sig)) in proposals {
-                            shares
-                                .entry(*id)
-                                .or_default()
-                                .insert(t.clone(), sig.clone());
-                        }
-                    }
-                    _ => {
-                        let c = count.candidates.entry(candidate).or_default();
-                        *c += 1;
-                    }
+                let existing_vote = votes_by_honest_voter
+                    .entry(unpacked_vote.voter)
+                    .or_insert_with(|| unpacked_vote.clone());
+                if unpacked_vote.supersedes(existing_vote) {
+                    *existing_vote = unpacked_vote.clone();
                 }
             }
         }
 
+        // We always include voters in the voter set (even if they are faulty)
+        // so that we have an accurate reading of who has contributed votes.
+        // This is done to to aid in split-vote detection
+        count.voters.extend(votes_by_honest_voter.keys().copied());
+        count.voters.extend(faulty.iter().copied());
+
+        for vote in votes_by_honest_voter.into_values() {
+            let candidate = vote.candidate();
+
+            if let Ballot::SuperMajority { proposals, .. } = &vote.vote.ballot {
+                let sm_count = count.super_majorities.entry(candidate.clone()).or_default();
+
+                sm_count.count += 1;
+
+                for (t, (id, sig)) in proposals {
+                    sm_count
+                        .proposals
+                        .entry(t.clone())
+                        .or_default()
+                        .insert(*id as u64, sig.clone());
+                }
+            }
+
+            let c = count.candidates.entry(candidate).or_default();
+            *c += 1;
+        }
+
         count
     }
 
@@ -141,19 +172,18 @@ impl<T: Proposition> VoteCount<T> {
             .iter()
             .map(|(candidates, c)| (candidates, *c))
             .chain(
-                self.super_majorities
-                    .iter()
-                    .map(|(candidates, shares)| (candidates, shares.len())),
+                self.super_majority_with_most_votes()
+                    .map(|(candidates, sm_count)| (candidates, sm_count.count)),
             )
             .max_by_key(|(_, c)| *c)
     }
 
     pub fn super_majority_with_most_votes(
         &self,
-    ) -> Option<(&Candidate<T>, &SignatureSharesByVoter<T>)> {
+    ) -> Option<(&Candidate<T>, &SuperMajorityCount<T>)> {
         self.super_majorities
             .iter()
-            .max_by_key(|(_, shares)| shares.len())
+            .max_by_key(|(_, sm_count)| sm_count.count)
     }
 }
 
@@ -197,13 +227,6 @@ impl<T: Proposition> Debug for Vote<T> {
 }
 
 impl<T: Proposition> Vote<T> {
-    pub fn candidate(&self) -> Candidate<T> {
-        Candidate {
-            proposals: self.proposals(),
-            faulty: self.known_faulty(),
-        }
-    }
-
     pub fn is_super_majority_ballot(&self) -> bool {
         matches!(self.ballot, Ballot::SuperMajority { .. })
     }
@@ -212,12 +235,12 @@ impl<T: Proposition> Vote<T> {
         Ok(bincode::serialize(&self)?)
     }
 
-    pub fn known_faulty(&self) -> BTreeSet<NodeId> {
+    pub fn faulty_ids(&self) -> BTreeSet<NodeId> {
         BTreeSet::from_iter(self.faults.iter().map(Fault::voter_at_fault))
     }
 
     pub fn proposals(&self) -> BTreeSet<T> {
-        self.proposals_with_known_faults(&self.known_faulty())
+        self.proposals_with_known_faults(&self.faulty_ids())
     }
 
     pub fn proposals_with_known_faults(&self, known_faulty: &BTreeSet<NodeId>) -> BTreeSet<T> {
@@ -245,6 +268,19 @@ impl<T: Proposition> Debug for SignedVote<T> {
 }
 
 impl<T: Proposition> SignedVote<T> {
+    pub fn candidate(&self) -> Candidate<T> {
+        match &self.vote.ballot {
+            Ballot::SuperMajority { votes, .. } => VoteCount::count(votes, &self.vote.faulty_ids())
+                .candidate_with_most_votes()
+                .map(|(candidate, _)| candidate.clone())
+                .unwrap_or_default(),
+            _ => Candidate {
+                proposals: self.proposals(),
+                faulty: self.vote.faulty_ids(),
+            },
+        }
+    }
+
     pub fn validate_signature(&self, voters: &PublicKeySet) -> Result<()> {
         crate::verify_sig_share(&self.vote, &self.sig, self.voter, voters)
     }
@@ -263,12 +299,12 @@ impl<T: Proposition> SignedVote<T> {
     }
 
     pub fn supersedes(&self, other: &Self) -> bool {
-        let our_known_faulty = self.vote.known_faulty();
-        let other_known_faulty = other.vote.known_faulty();
+        let our_faulty = self.vote.faulty_ids();
+        let other_faulty = other.vote.faulty_ids();
 
         if (&self.voter, self.vote.gen, &self.vote.ballot)
             == (&other.voter, other.vote.gen, &other.vote.ballot)
-            && our_known_faulty.is_superset(&other_known_faulty)
+            && our_faulty.is_superset(&other_faulty)
         {
             true
         } else {
@@ -281,11 +317,7 @@ impl<T: Proposition> SignedVote<T> {
         }
     }
 
-    pub fn strict_supersedes(&self, signed_vote: &Self) -> bool {
-        self != signed_vote && self.supersedes(signed_vote)
-    }
-
     pub fn vote_count(&self) -> VoteCount<T> {
-        VoteCount::count([self], &self.vote.known_faulty())
+        VoteCount::count([self], &self.vote.faulty_ids())
     }
 }
